from felis.currying import curry
from felis.typing import SupportsRichComparison

__all__ = ["different", "greater_than", "greater_than_or_same_as", "less_than", "less_than_or_same_as", "same"]


@curry
def less_than[T: SupportsRichComparison](first: T, second: T) -> bool:
    return first < second


@curry
def less_than_or_same_as[T: SupportsRichComparison](first: T, second: T) -> bool:
    return first <= second


@curry
def same[T: SupportsRichComparison](first: T, second: T) -> bool:
    return first == second


@curry
def different[T: SupportsRichComparison](first: T, second: T) -> bool:
    return first != second


@curry
def greater_than_or_same_as[T: SupportsRichComparison](first: T, second: T) -> bool:
    return first >= second


@curry
def greater_than[T: SupportsRichComparison](first: T, second: T) -> bool:
    return first > second



from collections.abc import Callable
from typing import Any

from felis.currying import curry

__all__ = ["List", "apply", "join"]


List = list


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [From : *] -> [To : *] -> M (List (From -> To)) -> M (List From) -> M (List To)
@curry
@curry
@curry
def apply(
    m_list_value: Any,
    m_list_function: Any,
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    return m_bind(m_list_function)(
        lambda list_function: m_bind(m_list_value)(lambda list_value: m_identity([function(value) for function in list_function for value in list_value])),
    )


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> M (List (M (List T))) -> M (List T)
@curry
@curry
def join(m_list_m_list_value: Any, m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]], m_identity: Callable[[Any], Any]) -> Any:
    def list_m_list_binder(list_m_list_value: List[Any]) -> Any:
        m_list = m_identity([])
        for m_list_value in list_m_list_value:
            m_list = m_bind(m_list_value)(lambda addend, current_m_list=m_list: m_bind(current_m_list)(lambda augend: m_identity(augend + addend)))
        return m_list

    return m_bind(m_list_m_list_value)(list_m_list_binder)



import builtins
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import felis.identity
import felis.order
from felis import applicative, function, list_t, monad
from felis.currying import curry
from felis.list_t import List
from felis.order import Order
from felis.predicate import Predicate

__all__ = [
    "List",
    "add",
    "append",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "filter",
    "fold",
    "guard",
    "identity",
    "join",
    "lift2",
    "map",
    "neutral",
    "range",
    "sort",
    "take_after",
    "take_before",
    "traverse",
    "when",
]


# [T : *] -> List T
neutral: list[Any] = []


@curry
def append[T](list: List[T], value: T) -> List[T]:
    return [*list, value]


@curry
def add[T](augend: List[T], addend: List[T]) -> List[T]:
    return augend + addend


@curry
def map[From, To](list_value: List[From], function: Callable[[From], To]) -> List[To]:
    return [function(value) for value in list_value]


def identity[T](value: T) -> List[T]:
    return [value]


if TYPE_CHECKING:

    @curry
    def apply[From, To](list_value: List[From], list_function: List[Callable[[From], To]]) -> List[To]: ...

else:
    apply = list_t.apply(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](second: List[Second], first: List[First], function: Callable[[First], Callable[[Second], Result]]) -> List[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, list_none: List[None]) -> List[None]: ...

else:
    when = applicative.when(identity)


@curry
@curry
def fold[A, T](list_value: List[T], function: Callable[[T], Callable[[A], A]], accumulator: A) -> A:
    for value in list_value:
        accumulator = function(value)(accumulator)
    return accumulator


@curry
@curry
def fold_right[A, T](list_value: List[T], function: Callable[[T], Callable[[A], A]], accumulator: A) -> A:
    if list_value:
        return function(list_value[0])(fold_right(accumulator)(function)(list_value[1:]))
    return accumulator


@curry
@curry
def fold_left[A, T](list_value: List[T], function: Callable[[A], Callable[[T], A]], accumulator: A) -> A:
    if list_value:
        return fold_left(function(accumulator)(list_value[0]))(function)(list_value[1:])
    return accumulator


# [A : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> A From -> A To) ->
# ([T : *] -> T -> A T) ->
# [From : *] -> [To : *] -> (From -> A To) -> List From -> A (List To)
@curry
@curry
def traverse[From](
    function: Callable[[From], Any],
    a_lift2: Callable[[Callable[[Any], Callable[[Any], Any]]], Callable[[Any], Callable[[Any], Any]]],
    a_identity: Callable[[Any], Any],
) -> Callable[[List[From]], Any]:
    return fold(a_identity(neutral))(felis.identity.compose(a_lift2(append))(function))


if TYPE_CHECKING:

    def join[T](list_list_value: List[List[T]]) -> List[T]: ...

else:
    join = list_t.join(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](list_value: List[From], function: Callable[[From], List[To]]) -> List[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](value: From, first: Callable[[From], List[Intermediate]], second: Callable[[Intermediate], List[To]]) -> List[To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    def guard(bool: bool) -> List[None]: ...

else:
    guard = monad.guard(neutral)(identity)


@curry
def filter[T](list: List[T], predicate: Predicate[T]) -> List[T]:
    return [value for value in list if predicate(value)]


@curry
def sort[T](list: List[T], order: Order[T]) -> List[T]:
    return sorted(list, key=felis.order.rich_comparison(order))


@curry
def range(stop: int, start: int) -> List[int]:
    return list(builtins.range(start, stop))



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from felis import applicative, monad
from felis.currying import curry
from felis.function import flip_, identity_

__all__ = [
    "Identity",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "inject",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type Identity[T] = T


@curry
def map[From, To](identity_value: Identity[From], function: Callable[[From], To]) -> Identity[To]:
    return function(identity_value)


def identity[T](value: T) -> Identity[T]:
    return value


@curry
def apply[From, To](identity_value: Identity[From], identity_function: Identity[Callable[[From], To]]) -> Identity[To]:
    return identity_function(identity_value)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: Identity[Second],
        first: Identity[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Identity[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(flip_.flip(identity_.identity))


discard_after = lift2(identity_.identity)


take_before = flip_.flip(discard_after)


discard_before = flip_.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, identity_none: Identity[None]) -> Identity[None]: ...

else:
    when = applicative.when(identity)


# [M : * -> *] -> [T : *] -> Identity (M T) -> M T
def inject(identity_m_identity_value: Identity[Any]) -> Any:
    return identity_m_identity_value


join = inject


if TYPE_CHECKING:

    @curry
    def bound[From, To](identity_value: Identity[From], function: Callable[[From], Identity[To]]) -> Identity[To]: ...

else:
    bound = monad.bound(map)(join)


bind = flip_.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], Identity[Intermediate]],
        second: Callable[[Intermediate], Identity[To]],
    ) -> Identity[To]: ...

else:
    compose = monad.compose(bind)



from collections.abc import Callable
from typing import Any

from felis.currying import curry
from felis.lazy import Lazy

__all__ = [
    "add",
    "apply",
    "identity",
    "join",
    "map",
    "neutral",
    "reversed_apply",
    "reversed_join",
    "run",
]


# [S : *] -> [M : * -> *] -> ([From : *] -> [To : *] -> (From -> To) -> M From -> M To) -> [T : *] -> StateT S M T -> S -> M T
@curry
@curry
def run[S](state: S, state_value: Callable[[S], Any], m_map: Callable[[Callable[[Any], Any]], Callable[[Any], Any]]) -> Any:
    def value_mapper(value_and_state: tuple[Any, S]) -> Any:
        value, _ = value_and_state
        return value

    return m_map(value_mapper)(state_value(state))


# [S : *] -> [M : * -> *] -> ([T : *] -> M T) -> [T : *] -> StateT S M T
@curry
def neutral(state: Any, m_neutral: Any) -> Any:
    return m_neutral


# [S : *] -> [M : * -> *] -> ([T : *] -> M T -> M T -> M T) -> [T : *] -> StateT S M T -> StateT S M T -> StateT S M T
@curry
@curry
@curry
def add[S](state: S, first: Callable[[S], Any], second: Callable[[S], Any], m_add: Callable[[Any], Any]) -> Any:
    return m_add(second(state))(first(state))


# [S : *] -> [M : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> M From -> T To) ->
# [From : *] -> [To : *] -> (From -> To) -> StateT S M From -> StateT S M To
@curry
@curry
@curry
def map[S](
    state: S,
    state_value: Callable[[S], Any],
    function: Callable[[Any], Any],
    m_map: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
) -> Any:
    def value_mapper(value_and_new_state: tuple[Any, S]) -> tuple[Any, S]:
        value, new_state = value_and_new_state
        return function(value), new_state

    return m_map(value_mapper)(state_value(state))


# [S : *] -> [M : * -> *] -> ([T : *] -> T -> M T) -> [T : *] -> T -> StateT S M T
@curry
@curry
def identity(state: Any, value: Any, m_identity: Callable[[Any], Any]) -> Any:
    return m_identity((value, state))


# [S : *] -> [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [From : *] -> [To : *] -> StateT S M (From -> To) -> StateT S M From -> StateT S M To
@curry
@curry
@curry
@curry
def apply[S](
    state: S,
    state_value: Callable[[S], Any],
    state_function: Callable[[S], Any],
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def function_binder(function_and_intermediate_state: tuple[Callable[[Any], Any], S]) -> Any:
        function, intermediate_state = function_and_intermediate_state

        def value_binder(value_and_new_state: tuple[Any, S]) -> Any:
            value, new_state = value_and_new_state
            return m_identity((function(value), new_state))

        return m_bind(state_value(intermediate_state))(value_binder)

    return m_bind(state_function(state))(function_binder)


# [S : *] -> [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> StateT S M (StateT S M T) -> StateT S M T
@curry
@curry
@curry
def join[S](
    state: S,
    state_state_value: Callable[[S], Any],
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def state_value_binder(state_value_and_intermediate_state: tuple[Any, S]) -> Any:
        state_value, intermediate_state = state_value_and_intermediate_state

        def value_binder(value_and_new_state: tuple[Any, S]) -> Any:
            value, new_state = value_and_new_state
            return m_identity((value, new_state))

        return m_bind(state_value(intermediate_state))(value_binder)

    return m_bind(state_state_value(state))(state_value_binder)


# [S : *] -> [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [From : *] -> [To : *] -> ReversedStateT S M (From -> To) -> ReversedStateT S M From -> ReversedStateT S M To
@curry
@curry
@curry
@curry
def reversed_apply[S](
    state: Lazy[S],
    reversed_state_value: Callable[[Lazy[S]], Any],
    reversed_state_function: Callable[[Lazy[S]], Any],
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def function_binder(function_and_new_state: tuple[Callable[[Any], Any], Lazy[S]], /) -> Any:
        nonlocal state
        function, new_state = function_and_new_state

        def value_binder(value_and_state: tuple[Any, Lazy[S]], /) -> Any:
            nonlocal state
            value, state = value_and_state
            return m_identity((function(value), new_state))

        return m_bind(reversed_state_value(new_state))(value_binder)

    return m_bind(reversed_state_function(lambda: state()))(function_binder)


# [S : *] -> [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> ReversedStateT S M (ReversedStateT S M T) -> ReversedStateT S M T
@curry
@curry
@curry
def reversed_join[S](
    state: Lazy[S],
    reversed_state_reversed_state_value: Callable[[Lazy[S]], Any],
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def reversed_state_value_binder(reversed_state_value_and_new_state: tuple[Any, S]) -> Any:
        nonlocal state
        reversed_state_value, new_state = reversed_state_value_and_new_state

        def value_binder(value_and_state: tuple[Any, Lazy[S]]) -> Any:
            nonlocal state
            value, state = value_and_state
            return m_identity((value, new_state))

        return m_bind(reversed_state_value(state))(value_binder)

    return m_bind(reversed_state_reversed_state_value(lambda: state()))(reversed_state_value_binder)



from typing import NewType

import felis.predicate.bool
from felis.constrained import smart_constructor

__all__ = ["NonEmpty", "non_empty"]


NonEmpty = NewType("NonEmpty", str)

non_empty = smart_constructor(NonEmpty)(felis.predicate.bool.truthy)



from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import felis.function
from felis.currying import curry

__all__ = ["Bind", "Free", "Pure", "apply", "identity", "join", "map"]


@dataclass(frozen=True)
class Pure[T]:
    value: T


# [F : * -> *] -> [T : *] -> F (Free F T) -> Free F T
@dataclass(frozen=True)
class Bind:
    f_free_value: Any


type Free[T] = Pure[T] | Bind


# [F : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> F From -> F To) ->
# [From : *] -> [To : *] -> (From -> To) -> Free F From -> Free F To
@curry
@curry
def map[From, To](free_value: Free[From], function: Callable[[From], To], f_map: Callable[[Any], Callable[[Any], Any]]) -> Free[To]:
    match free_value:
        case Pure(value):
            return Pure(function(value))
        case Bind(f_free_value):
            return Bind(f_map(map(f_map)(function))(f_free_value))


if TYPE_CHECKING:
    # [F : * -> *] -> [T : *] -> T -> Free F T
    identity: Free[Any]
else:
    identity = Pure


# [F : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> F From -> F To) ->
# [From : *] -> [To : *] -> Free F (From -> To) -> Free F From -> Free F To
@curry
@curry
def apply[From, To](free_value: Free[From], free_function: Free[Callable[[From], To]], f_map: Callable[[Any], Callable[[Any], Any]]) -> Free[To]:
    match free_function:
        case Pure(function):
            return map(f_map)(function)(free_value)
        case Bind(f_free_function):
            return Bind(f_map(felis.function.flip(apply(f_map))(free_value))(f_free_function))


# [F : * -> *] -> ([From : *] -> [To : *] -> (From -> To) -> F From -> F To) -> [T : *] -> Free F (Free F T) -> Free F T
@curry
def join[T](free_free_value: Free[Free[T]], f_map: Callable[[Any], Callable[[Any], Any]]) -> Free[T]:
    match free_free_value:
        case Pure(free_value):
            return free_value
        case Bind(f_free_free_value):
            return Bind(f_map(join(f_map))(f_free_free_value))



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, monad
from felis.currying import curry
from felis.function.flip_ import flip
from felis.function.identity_ import identity

__all__ = [
    "Function",
    "add",
    "add2",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "flip",
    "identity",
    "join",
    "lift2",
    "map",
    "map2",
    "neutral",
    "neutral2",
    "take_after",
    "take_before",
    "when",
]


type Function[From, To] = Callable[[From], To]


@curry
def neutral[M](_: object, m_neutral: M) -> M:
    return m_neutral


neutral2 = felis.identity.compose(neutral)(neutral)


@curry
@curry
@curry
def add[M, T](value: T, augend: Function[T, M], addend: Function[T, M], m_add: Callable[[M], Callable[[M], M]]) -> M:
    return m_add(addend(value))(augend(value))


add2 = felis.identity.compose(add)(add)


@curry
@curry
def map[T, From, To](value: T, function_value: Function[T, From], function: Callable[[From], To]) -> To:
    return function(function_value(value))


map2 = felis.identity.compose(map)(map)


@curry
@curry
def apply[T, From, To](value: T, function_value: Function[T, From], function: Function[T, Callable[[From], To]]) -> To:
    return function(value)(function_value(value))


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[T, First, Second, Result](
        second: Function[T, Second],
        first: Function[T, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Function[T, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(flip(identity))


discard_after = lift2(identity)


take_before = flip(discard_after)


discard_before = flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[T](bool: bool, function_none: Function[T, None]) -> Function[T, None]: ...

else:
    when = applicative.when(identity)


@curry
def join[From, To](value: From, function_function_value: Function[From, Function[From, To]]) -> To:
    return function_function_value(value)(value)


if TYPE_CHECKING:

    @curry
    def bound[T, From, To](either_value: Function[T, From], function: Callable[[From], Function[T, To]]) -> Function[T, To]: ...

else:
    bound = monad.bound(map)(join)


bind = flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[T, From, Intermediate, To](
        value: From,
        first: Callable[[From], Function[T, Intermediate]],
        second: Callable[[Intermediate], Function[T, To]],
    ) -> Function[T, To]: ...

else:
    compose = monad.compose(bind)



from felis.currying import curry
from felis.predicate import Predicate, both, comparison
from felis.typing import SupportsRichComparison

__all__ = ["exclusive", "exclusive_inclusive", "inclusive", "inclusive_exclusive"]


@curry
def inclusive[T: SupportsRichComparison](maximum: T, minimum: T) -> Predicate[T]:
    return both(comparison.greater_than_or_same_as(minimum))(comparison.less_than_or_same_as(maximum))


@curry
def inclusive_exclusive[T: SupportsRichComparison](maximum: T, minimum: T) -> Predicate[T]:
    return both(comparison.greater_than_or_same_as(minimum))(comparison.less_than(maximum))


@curry
def exclusive_inclusive[T: SupportsRichComparison](maximum: T, minimum: T) -> Predicate[T]:
    return both(comparison.greater_than(minimum))(comparison.less_than_or_same_as(maximum))


@curry
def exclusive[T: SupportsRichComparison](maximum: T, minimum: T) -> Predicate[T]:
    return both(comparison.greater_than(minimum))(comparison.less_than(maximum))



from collections.abc import Callable

from felis import Option, Predicate, option
from felis.currying import curry

__all__ = ["smart_constructor"]


@curry
@curry
def smart_constructor[From, To](value: From, predicate: Predicate[From], constructor: Callable[[From], To]) -> Option[To]:
    return option.Some(constructor(value)) if predicate(value) else None



from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from felis.currying import curry

__all__ = ["Option", "Some", "add", "apply", "default", "join", "neutral"]


type Option[T] = None | Some[T]


@dataclass(frozen=True)
class Some[T]:
    value: T


# [M : * -> *] -> ([T : *] -> T -> M T) -> [T : *] -> M (Option T)
def neutral(m_identity: Callable[[Any], Any]) -> Any:
    return m_identity(None)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> M (Option T) -> M (Option T) -> M (Option T)
@curry
@curry
@curry
def add(m_augend: Any, m_addend: Any, m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]], m_identity: Callable[[Any], Any]) -> Any:
    def augend_binder(augend: Option[Any]) -> Any:
        match augend:
            case None:
                return m_addend
            case Some(value):
                return m_identity(Some(value))

    return m_bind(m_augend)(augend_binder)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [From : *] -> [To : *] -> M (Option (From -> To)) -> M (Option From) -> M (Option To)
@curry
@curry
@curry
def apply(
    m_either_value: Any,
    m_either_function: Any,
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def option_function_binder(option_function: Option[Callable[[Any], Any]]) -> Any:
        match option_function:
            case None:
                return m_identity(None)
            case Some(function):

                def option_value_binder(option_value: Option[Any]) -> Any:
                    match option_value:
                        case None:
                            return m_identity(None)
                        case Some(value):
                            return m_identity(Some(function(value)))

                return m_bind(m_either_value)(option_value_binder)

    return m_bind(m_either_function)(option_function_binder)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> M (Option (M (Option T))) -> M (Option T)
@curry
@curry
def join(m_option_m_option_value: Any, m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]], m_identity: Callable[[Any], Any]) -> Any:
    def binder(option_m_option_value: Option[Any]) -> Any:
        match option_m_option_value:
            case None:
                return m_identity(None)
            case Some(m_option_value):
                return m_option_value

    return m_bind(m_option_m_option_value)(binder)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> M T -> M (Option T) -> M T
@curry
@curry
@curry
def default(
    m_option_value: Any,
    default_value: Any,
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def binder(option_value: Option[Any]) -> Any:
        match option_value:
            case None:
                return default_value
            case Some(value):
                return m_identity(value)

    return m_bind(m_option_value)(binder)



from collections.abc import Callable
from typing import Any

from felis import option
from felis.currying import curry
from felis.option import Option

__all__ = ["Dict", "add", "get", "map", "neutral", "set"]


Dict = dict


# [K : *] -> [V : *] -> dict K V
neutral: dict[Any, Any] = {}


@curry
@curry
def set[K, V](dict: Dict[K, V], value: V, key: K) -> Dict[K, V]:
    return {**dict, key: value}


@curry
def get[K, V](dict: Dict[K, V], key: K) -> Option[V]:
    return option.Some(dict[key]) if key in dict else None


@curry
def add[K, V](augend: Dict[K, V], addend: Dict[K, V]) -> Dict[K, V]:
    return {**augend, **addend}


@curry
def map[K, From, To](dict_value: Dict[K, From], function: Callable[[From], To]) -> Dict[K, To]:
    return {key: function(value) for key, value in dict_value.items()}



import builtins
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any

import felis.identity
import felis.order
from felis import applicative, function, monad
from felis.currying import curry
from felis.predicate import Predicate

__all__ = [
    "Iterable",
    "add",
    "append",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "filter",
    "fold",
    "guard",
    "identity",
    "join",
    "lift2",
    "map",
    "neutral",
    "range",
    "take_after",
    "take_before",
    "traverse",
    "when",
]


# [T : *] -> Iterable T
neutral: Iterable[Any] = ()


@curry
def append[T](iterable: Iterable[T], value: T) -> Iterable[T]:
    yield from iterable
    yield value


@curry
def add[T](augend: Iterable[T], addend: Iterable[T]) -> Iterable[T]:
    yield from augend
    yield from addend


@curry
def map[From, To](iterable_value: Iterable[From], function: Callable[[From], To]) -> Iterable[To]:
    for value in iterable_value:
        yield function(value)


def identity[T](value: T) -> Iterable[T]:
    yield value


@curry
def apply[From, To](iterable_value: Iterable[From], iterable_function: Iterable[Callable[[From], To]]) -> Iterable[To]:
    for function in iterable_function:
        for value in iterable_value:
            yield function(value)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: Iterable[Second],
        first: Iterable[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Iterable[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, iterable_none: Iterable[None]) -> Iterable[None]: ...

else:
    when = applicative.when(identity)


@curry
@curry
def fold[A, T](iterable_value: Iterable[T], function: Callable[[T], Callable[[A], A]], accumulator: A) -> A:
    for value in iterable_value:
        accumulator = function(value)(accumulator)
    return accumulator


# [A : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> A From -> A To) ->
# ([T : *] -> T -> A T) ->
# [From : *] -> [To : *] -> (From -> A To) -> Iterable From -> A (Iterable To)
@curry
@curry
def traverse[From](
    function: Callable[[From], Any],
    a_lift2: Callable[[Callable[[Any], Callable[[Any], Any]]], Callable[[Any], Callable[[Any], Any]]],
    a_identity: Callable[[Any], Any],
) -> Callable[[Iterable[From]], Any]:
    return fold(a_identity(neutral))(felis.identity.compose(a_lift2(append))(function))


if TYPE_CHECKING:

    def join[T](iterable_iterable_value: Iterable[Iterable[T]]) -> Iterable[T]: ...

else:
    join = fold(neutral)(add)


if TYPE_CHECKING:

    @curry
    def bound[From, To](iterable_value: Iterable[From], function: Callable[[From], Iterable[To]]) -> Iterable[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], Iterable[Intermediate]],
        second: Callable[[Intermediate], Iterable[To]],
    ) -> Iterable[To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    def guard(bool: bool) -> Iterable[None]: ...

else:
    guard = monad.guard(neutral)(identity)


@curry
def filter[T](iterable: Iterable[T], predicate: Predicate[T]) -> Iterable[T]:
    for value in iterable:
        if predicate(value):
            yield value


@curry
def range(stop: int, start: int) -> Iterable[int]:
    return builtins.range(start, stop)



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import felis.identity
from felis import applicative, function, monad, option_t
from felis.currying import curry
from felis.option_t import Option, Some

__all__ = [
    "Option",
    "Some",
    "add",
    "apply",
    "bind",
    "bound",
    "compose",
    "default",
    "discard_after",
    "discard_before",
    "fold",
    "guard",
    "identity",
    "join",
    "lift2",
    "map",
    "neutral",
    "take_after",
    "take_before",
    "traverse",
    "when",
]


if TYPE_CHECKING:
    neutral: Option[Any]
else:
    neutral = option_t.neutral(felis.identity.identity)


if TYPE_CHECKING:

    @curry
    def add[T](augend: Option[T], addend: Option[T]) -> Option[T]: ...

else:
    add = option_t.add(felis.identity.identity)(felis.identity.bind)


@curry
def map[From, To](option_value: Option[From], function: Callable[[From], To]) -> Option[To]:
    match option_value:
        case None:
            return None
        case Some(value):
            return Some(function(value))


if TYPE_CHECKING:

    def identity[T](value: T, /) -> Option[T]: ...

else:
    identity = Some


if TYPE_CHECKING:

    @curry
    def apply[From, To](option_value: Option[From], option_function: Option[Callable[[From], To]]) -> Option[To]: ...

else:
    apply = option_t.apply(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: Option[Second],
        first: Option[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Option[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, option_none: Option[None]) -> Option[None]: ...

else:
    when = applicative.when(identity)


@curry
@curry
def fold[A, T](option_value: Option[T], function: Callable[[T], Callable[[A], A]], accumulator: A) -> A:
    match option_value:
        case None:
            return accumulator
        case Some(value):
            return function(value)(accumulator)


# [A : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> A From -> A To) ->
# ([T : *] -> T -> A T) ->
# [From : *] -> [To : *] -> (From -> A To) -> Option From -> A (Option To)
@curry
@curry
@curry
def traverse[From](
    option_value: Option[From],
    function: Callable[[From], Any],
    a_identity: Callable[[Any], Any],
    a_map: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
) -> Any:
    match option_value:
        case None:
            return a_identity(neutral)
        case Some(value):
            return a_map(identity)(function(value))


if TYPE_CHECKING:

    def join[From, To](option_option_value: Option[Option[From]]) -> Option[To]: ...

else:
    join = option_t.join(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](option_value: Option[From], function: Callable[[From], Option[To]]) -> Option[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], Option[Intermediate]],
        second: Callable[[Intermediate], Option[To]],
    ) -> Option[To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    def guard(bool: bool) -> Option[None]: ...

else:
    guard = monad.guard(neutral)(identity)


if TYPE_CHECKING:

    @curry
    def default[T](option_value: Option[T], default_value: T) -> T: ...

else:
    default = option_t.default(felis.identity.identity)(felis.identity.bind)



from collections.abc import Callable
from typing import Any

from felis.currying import curry

__all__ = ["lift2", "when"]


# [A : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> A From -> A To) ->
# ([From : *] -> [To : *] -> A (From -> To) -> A From -> A To) ->
# [First : *] -> [Second : *] -> [Result : *] -> (First -> Second -> Result) -> A First -> A Second -> A Result
@curry
@curry
@curry
def lift2(
    first: Any,
    function: Callable[[Any], Callable[[Any], Any]],
    apply: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
    map: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
) -> Callable[[Any], Any]:
    return apply(map(function)(first))


# [A : * -> *] -> ([T : *] -> T -> A T) -> bool -> A None -> A None
@curry
@curry
def when(a_none: Any, bool: bool, identity: Callable[[Any], Any]) -> Any:
    return a_none if bool else identity(None)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from felis import applicative, dict, free_t, function, monad
from felis.currying import curry
from felis.free_t import Pure

__all__ = [
    "Bind",
    "FreeDict",
    "Pure",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


if TYPE_CHECKING:

    @dataclass(frozen=True)
    class Bind[K, T]:
        f_free_dict_value: dict.Dict[K, FreeDict[K, T]]

else:
    from felis.free_t import Bind


if TYPE_CHECKING:
    type FreeDict[K, T] = Pure[T] | Bind[K, T]
else:
    from felis.free_t import Free as FreeDict


if TYPE_CHECKING:

    @curry
    def map[K, From, To](free_dict_value: FreeDict[K, From], function: Callable[[From], To]) -> FreeDict[K, To]: ...

else:
    map = free_t.map(dict.map)


if TYPE_CHECKING:
    # [K : *] -> [T : *] -> T -> FreeDict K T
    identity: FreeDict[Any, Any]
else:
    from felis.free_t import identity


if TYPE_CHECKING:

    @curry
    def apply[K, From, To](free_dict_value: FreeDict[K, From], free_dict_function: FreeDict[K, Callable[[From], To]]) -> FreeDict[K, To]: ...

else:
    apply = free_t.apply(dict.map)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[K, First, Second, Result](
        second: FreeDict[K, Second],
        first: FreeDict[K, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> FreeDict[K, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[K](bool: bool, free_dict_none: FreeDict[K, None]) -> FreeDict[K, None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[K, T](free_dict_free_dict_value: FreeDict[K, FreeDict[K, T]], /) -> FreeDict[K, T]: ...

else:
    join = free_t.join(dict.map)


if TYPE_CHECKING:

    @curry
    def bound[K, From, To](free_dict_value: FreeDict[K, From], function: Callable[[From], FreeDict[K, To]]) -> FreeDict[K, To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[K, From, Intermediate, To](
        value: From,
        first: Callable[[From], FreeDict[K, Intermediate]],
        second: Callable[[Intermediate], FreeDict[K, To]],
    ) -> FreeDict[K, To]: ...

else:
    compose = monad.compose(bind)



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, coroutine, function, monad, option, option_t
from felis.coroutine import Coroutine
from felis.currying import curry
from felis.option import Option

__all__ = [
    "CoroutineOption",
    "add",
    "apply",
    "bind",
    "bound",
    "compose",
    "default",
    "discard_after",
    "discard_before",
    "guard",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type CoroutineOption[T] = Coroutine[Option[T]]


if TYPE_CHECKING:

    @curry
    def add[T](coroutine_option_augend: CoroutineOption[T], coroutine_option_addend: CoroutineOption[T]) -> CoroutineOption[T]: ...

else:
    add = option_t.add(coroutine.identity)(coroutine.bind)


if TYPE_CHECKING:

    @curry
    def map[From, To](coroutine_option_value: CoroutineOption[From], function: Callable[[From], To]) -> CoroutineOption[To]: ...

else:
    map = felis.identity.compose(coroutine.map)(option.map)


identity = felis.identity.compose(coroutine.identity)(option.identity)


apply = coroutine.lift2(option.apply)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: CoroutineOption[Second],
        first: CoroutineOption[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> CoroutineOption[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, coroutine_option_none: CoroutineOption[None]) -> CoroutineOption[None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T](coroutine_option_coroutine_option_value: CoroutineOption[CoroutineOption[T]]) -> CoroutineOption[T]: ...

else:
    join = option_t.join(coroutine.identity)(coroutine.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](coroutine_option_value: CoroutineOption[From], function: Callable[[From], CoroutineOption[To]]) -> CoroutineOption[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], CoroutineOption[Intermediate]],
        second: Callable[[Intermediate], CoroutineOption[To]],
    ) -> CoroutineOption[To]: ...

else:
    compose = monad.compose(bind)


guard = felis.identity.compose(coroutine.identity)(option.guard)


if TYPE_CHECKING:

    @curry
    def default[T](coroutine_option_value: CoroutineOption[T], default_value: Coroutine[T]) -> Coroutine[T]: ...

else:
    default = option_t.default(coroutine.identity)(coroutine.bind)



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, function, lazy_coroutine, monad, option, option_t
from felis.currying import curry
from felis.lazy_coroutine import LazyCoroutine
from felis.option import Option

__all__ = [
    "LazyCoroutineOption",
    "add",
    "apply",
    "bind",
    "bound",
    "compose",
    "default",
    "discard_after",
    "discard_before",
    "guard",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type LazyCoroutineOption[T] = LazyCoroutine[Option[T]]


if TYPE_CHECKING:

    @curry
    def add[T](lazy_coroutine_option_augend: LazyCoroutineOption[T], lazy_coroutine_option_addend: LazyCoroutineOption[T]) -> LazyCoroutineOption[T]: ...

else:
    add = option_t.add(lazy_coroutine.identity)(lazy_coroutine.bind)


if TYPE_CHECKING:

    @curry
    def map[From, To](lazy_coroutine_option_value: LazyCoroutineOption[From], function: Callable[[From], To]) -> LazyCoroutineOption[To]: ...

else:
    map = felis.identity.compose(lazy_coroutine.map)(option.map)


identity = felis.identity.compose(lazy_coroutine.identity)(option.identity)


apply = lazy_coroutine.lift2(option.apply)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: LazyCoroutineOption[Second],
        first: LazyCoroutineOption[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> LazyCoroutineOption[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, lazy_coroutine_option_none: LazyCoroutineOption[None]) -> LazyCoroutineOption[None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T](lazy_coroutine_option_lazy_coroutine_option_value: LazyCoroutineOption[LazyCoroutineOption[T]]) -> LazyCoroutineOption[T]: ...

else:
    join = option_t.join(lazy_coroutine.identity)(lazy_coroutine.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](
        lazy_coroutine_option_value: LazyCoroutineOption[From],
        function: Callable[[From], LazyCoroutineOption[To]],
    ) -> LazyCoroutineOption[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], LazyCoroutineOption[Intermediate]],
        second: Callable[[Intermediate], LazyCoroutineOption[To]],
    ) -> LazyCoroutineOption[To]: ...

else:
    compose = monad.compose(bind)


guard = felis.identity.compose(lazy_coroutine.identity)(option.guard)


if TYPE_CHECKING:

    @curry
    def default[T](lazy_coroutine_option_value: LazyCoroutineOption[T], default_value: LazyCoroutine[T]) -> LazyCoroutine[T]: ...

else:
    default = option_t.default(lazy_coroutine.identity)(lazy_coroutine.bind)



from collections.abc import Callable
from typing import Any

from felis.currying import curry

__all__ = ["bound", "compose", "guard"]


# [M : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> M From -> M To) ->
# ([T : *] -> M (M T) -> M T) ->
# [From : *] -> [To : *] -> (From -> M To) -> M From -> M To
@curry
@curry
@curry
def bound(
    m_value: Any,
    function: Callable[[Any], Any],
    join: Callable[[Any], Any],
    map: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
) -> Any:
    return join(map(function)(m_value))


# [M : * -> *] ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [Intermediate : *] -> [To : *] -> (Intermediate -> M To) -> [From : *] -> (From -> M Intermediate) -> From -> M To
@curry
@curry
@curry
def compose[From](
    value: From,
    first: Callable[[From], Any],
    second: Callable[[Any], Any],
    bind: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
) -> Any:
    return bind(first(value))(second)


# [A : * -> *] -> ([T : *] -> A T) -> ([T : *] -> T -> A T) -> bool -> A None
@curry
@curry
def guard(bool: bool, identity: Callable[[Any], Any], neutral: Any) -> Any:
    return identity(None) if bool else neutral



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, function, lazy_coroutine_either, list, list_t, monad
from felis.currying import curry
from felis.lazy_coroutine_either import LazyCoroutineEither
from felis.list import List

__all__ = [
    "LazyCoroutineEitherList",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type LazyCoroutineEitherList[L, R] = LazyCoroutineEither[L, List[R]]


if TYPE_CHECKING:

    @curry
    def map[L, From, To](
        lazy_coroutine_either_list_value: LazyCoroutineEitherList[L, From],
        function: Callable[[From], To],
    ) -> LazyCoroutineEitherList[L, To]: ...

else:
    map = felis.identity.compose(lazy_coroutine_either.map)(list.map)


identity = felis.identity.compose(lazy_coroutine_either.identity)(list.identity)


apply = lazy_coroutine_either.lift2(list.apply)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[L, First, Second, Result](
        second: LazyCoroutineEitherList[L, Second],
        first: LazyCoroutineEitherList[L, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> LazyCoroutineEitherList[L, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[L](bool: bool, lazy_coroutine_either_list_none: LazyCoroutineEitherList[L, None]) -> LazyCoroutineEitherList[L, None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[L, R](
        lazy_coroutine_either_list_lazy_coroutine_either_list_value: LazyCoroutineEitherList[L, LazyCoroutineEitherList[L, R]],
    ) -> LazyCoroutineEitherList[L, R]: ...

else:
    join = list_t.join(lazy_coroutine_either.identity)(lazy_coroutine_either.bind)


if TYPE_CHECKING:

    @curry
    def bound[L, From, To](
        lazy_coroutine_either_list_value: LazyCoroutineEitherList[L, From],
        function: Callable[[From], LazyCoroutineEitherList[L, To]],
    ) -> LazyCoroutineEitherList[L, To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[L, From, Intermediate, To](
        value: From,
        first: Callable[[From], LazyCoroutineEitherList[L, Intermediate]],
        second: Callable[[Intermediate], LazyCoroutineEitherList[L, To]],
    ) -> LazyCoroutineEitherList[L, To]: ...

else:
    compose = monad.compose(bind)



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, either, either_t, function, lazy_coroutine, monad
from felis.currying import curry
from felis.either import Either
from felis.lazy_coroutine import LazyCoroutine

__all__ = [
    "LazyCoroutineEither",
    "add",
    "apply",
    "bind",
    "bound",
    "compose",
    "default",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type LazyCoroutineEither[L, R] = LazyCoroutine[Either[L, R]]


if TYPE_CHECKING:

    @curry
    def add[L, R](
        lazy_coroutine_either_augend: LazyCoroutineEither[L, R],
        lazy_coroutine_either_addend: LazyCoroutineEither[L, R],
    ) -> LazyCoroutineEither[L, R]: ...

else:
    add = either_t.add(lazy_coroutine.identity)(lazy_coroutine.bind)


if TYPE_CHECKING:

    @curry
    def map[L, From, To](lazy_coroutine_either_value: LazyCoroutineEither[L, From], function: Callable[[From], To]) -> LazyCoroutineEither[L, To]: ...

else:
    map = felis.identity.compose(lazy_coroutine.map)(either.map)


identity = felis.identity.compose(lazy_coroutine.identity)(either.identity)


apply = lazy_coroutine.lift2(either.apply)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[L, First, Second, Result](
        second: LazyCoroutineEither[L, Second],
        first: LazyCoroutineEither[L, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> LazyCoroutineEither[L, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[L](bool: bool, lazy_coroutine_either_none: LazyCoroutineEither[L, None]) -> LazyCoroutineEither[L, None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[L, R](lazy_coroutine_either_lazy_coroutine_either_value: LazyCoroutineEither[L, LazyCoroutineEither[L, R]]) -> LazyCoroutineEither[L, R]: ...

else:
    join = either_t.join(lazy_coroutine.identity)(lazy_coroutine.bind)


if TYPE_CHECKING:

    @curry
    def bound[L, From, To](
        lazy_coroutine_either_value: LazyCoroutineEither[L, From],
        function: Callable[[From], LazyCoroutineEither[L, To]],
    ) -> LazyCoroutineEither[L, To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[L, From, Intermediate, To](
        value: From,
        first: Callable[[From], LazyCoroutineEither[L, Intermediate]],
        second: Callable[[Intermediate], LazyCoroutineEither[L, To]],
    ) -> LazyCoroutineEither[L, To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    @curry
    def default[L, R](lazy_coroutine_either_value: LazyCoroutineEither[L, R], default_value: LazyCoroutine[R]) -> LazyCoroutine[R]: ...

else:
    default = either_t.default(lazy_coroutine.identity)(lazy_coroutine.bind)



from collections.abc import Callable
from typing import Any

from felis.currying import curry

__all__ = ["Lazy", "apply", "join"]


type Lazy[T] = Callable[[], T]


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [From : *] -> [To : *] -> Lazy (M (From -> To)) -> Lazy (M From) -> Lazy (M To)
@curry
@curry
@curry
@curry
def apply(
    lazy_m_value: Lazy[Any],
    lazy_m_function: Lazy[Any],
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Lazy[Any]:
    return lambda: m_bind(lazy_m_function())(lambda function: m_bind(lazy_m_value())(lambda value: m_identity(function(value))))


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [T : *] -> Lazy (M (Lazy (M T))) -> Lazy (M T)
@curry
@curry
def join(lazy_m_lazy_m_value: Lazy[Any], m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]], m_identity: Callable[[Any], Any]) -> Lazy[Any]:
    return lambda: m_bind(lazy_m_lazy_m_value())(lambda m_lazy_value: m_lazy_value())



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, function, monad, option, option_t
from felis.currying import curry
from felis.function import Function
from felis.option import Option

__all__ = [
    "FunctionOption",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "guard",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type FunctionOption[From, To] = Function[From, Option[To]]


if TYPE_CHECKING:

    @curry
    def map[T, From, To](function_option_value: FunctionOption[T, From], function: Callable[[From], To]) -> FunctionOption[T, To]: ...

else:
    map = felis.identity.compose(function.map)(option.map)


identity = felis.identity.compose(function.identity)(option.identity)


apply = function.lift2(option.apply)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[T, First, Second, Result](
        second: FunctionOption[T, Second],
        first: FunctionOption[T, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> FunctionOption[T, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[T](bool: bool, function_option_none: FunctionOption[T, None]) -> FunctionOption[T, None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T, From, To](function_option_function_option_value: FunctionOption[T, FunctionOption[T, From]]) -> FunctionOption[T, To]: ...

else:
    join = option_t.join(function.identity)(function.bind)


if TYPE_CHECKING:

    @curry
    def bound[T, From, To](
        function_option_value: FunctionOption[T, From],
        function: Callable[[From], FunctionOption[T, To]],
    ) -> FunctionOption[T, To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[T, From, Intermediate, To](
        value: From,
        first: Callable[[From], FunctionOption[T, Intermediate]],
        second: Callable[[Intermediate], FunctionOption[T, To]],
    ) -> FunctionOption[T, To]: ...

else:
    compose = monad.compose(bind)


guard = felis.identity.compose(function.identity)(option.guard)



import collections.abc
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from felis import applicative, function, monad
from felis.currying import curry

__all__ = [
    "Coroutine",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type Coroutine[T] = collections.abc.Coroutine[Any, Any, T]


@curry
async def map[From, To](coroutine_value: Coroutine[From], function: Callable[[From], To]) -> To:
    return function(await coroutine_value)


async def identity[T](value: T) -> T:
    return value


@curry
async def apply[From, To](coroutine_value: Coroutine[From], coroutine_function: Coroutine[Callable[[From], To]]) -> To:
    return await map(await coroutine_function)(coroutine_value)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: Coroutine[Second],
        first: Coroutine[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Coroutine[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, coroutine_none: Coroutine[None]) -> Coroutine[None]: ...

else:
    when = applicative.when(identity)


async def join[T](coroutine_coroutine_value: Coroutine[Coroutine[T]]) -> T:
    return await (await coroutine_coroutine_value)


if TYPE_CHECKING:

    @curry
    def bound[From, To](coroutine_value: Coroutine[From], function: Callable[[From], Coroutine[To]]) -> Coroutine[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], Coroutine[Intermediate]],
        second: Callable[[Intermediate], Coroutine[To]],
    ) -> Coroutine[To]: ...

else:
    compose = monad.compose(bind)



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, function, lazy_t, monad
from felis.currying import curry
from felis.lazy_t import Lazy

__all__ = [
    "Lazy",
    "add",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "run",
    "take_after",
    "take_before",
    "when",
]


def run[T](lazy_value: Lazy[T]) -> T:
    return lazy_value()


@curry
@curry
def add[M](first: Lazy[M], second: Lazy[M], m_add: Callable[[M], Callable[[M], M]]) -> Lazy[M]:
    return lambda: m_add(second())(first())


@curry
def map[From, To](lazy_value: Lazy[From], function: Callable[[From], To]) -> Lazy[To]:
    return lambda: function(lazy_value())


def identity[T](value: T) -> Lazy[T]:
    return lambda: value


if TYPE_CHECKING:

    @curry
    def apply[From, To](lazy_value: Lazy[From], lazy_function: Lazy[Callable[[From], To]]) -> Lazy[To]: ...

else:
    apply = lazy_t.apply(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](second: Lazy[Second], first: Lazy[First], function: Callable[[First], Callable[[Second], Result]]) -> Lazy[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, lazy_none: Lazy[None]) -> Lazy[None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T](lazy_lazy_value: Lazy[Lazy[T]]) -> Lazy[T]: ...

else:
    join = lazy_t.join(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](lazy_value: Lazy[From], function: Callable[[From], Lazy[To]]) -> Lazy[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](value: From, first: Callable[[From], Lazy[Intermediate]], second: Callable[[Intermediate], Lazy[To]]) -> Lazy[To]: ...

else:
    compose = monad.compose(bind)



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import felis.identity
from felis import applicative, either_t, function, monad
from felis.currying import curry
from felis.either_t import Either, Left, Right

__all__ = [
    "Either",
    "Left",
    "Right",
    "add",
    "apply",
    "bind",
    "bound",
    "catch",
    "compose",
    "default",
    "discard_after",
    "discard_before",
    "fold",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "traverse",
    "when",
]


if TYPE_CHECKING:

    @curry
    def add[L, R](augend: Either[L, R], addend: Either[L, R]) -> Either[L, R]: ...

else:
    add = either_t.add(felis.identity.identity)(felis.identity.bind)


@curry
def map[L, From, To](either_value: Either[L, From], function: Callable[[From], To]) -> Either[L, To]:
    match either_value:
        case Left(value):
            return Left(value)
        case Right(value):
            return Right(function(value))


if TYPE_CHECKING:
    # [L : *] -> [R : *] -> R -> Either L R
    def identity[R](value: R) -> Either[Any, R]: ...

else:
    identity = Right


if TYPE_CHECKING:

    @curry
    def apply[L, From, To](either_value: Either[L, From], either_function: Either[L, Callable[[From], To]]) -> Either[L, To]: ...

else:
    apply = either_t.apply(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[L, First, Second, Result](
        second: Either[L, Second],
        first: Either[L, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Either[L, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[L](bool: bool, either_none: Either[L, None]) -> Either[L, None]: ...

else:
    when = applicative.when(identity)


# [L : *] -> [A : *] -> A -> [R : *] -> (R -> A -> A) -> Either L R -> A
@curry
@curry
def fold[A, R](either_value: Either[Any, R], function: Callable[[R], Callable[[A], A]], accumulator: A) -> A:
    match either_value:
        case Left(value):
            return accumulator
        case Right(value):
            return function(value)(accumulator)


# [L : *] -> [A : * -> *] ->
# ([From : *] -> [To : *] -> (From -> To) -> A From -> A To) ->
# ([T : *] -> T -> A T) ->
# [From : *] -> [To : *] -> (From -> A To) -> Either L From -> A (Either L To)
@curry
@curry
@curry
def traverse[From](
    either_value: Either[Any, From],
    function: Callable[[From], Any],
    a_identity: Callable[[Any], Any],
    a_map: Callable[[Callable[[Any], Any]], Callable[[Any], Any]],
) -> Any:
    match either_value:
        case Left(value):
            return a_identity(Left(value))
        case Right(value):
            return a_map(identity)(function(value))


if TYPE_CHECKING:

    def join[L, R](either_either_value: Either[L, Either[L, R]], /) -> Either[L, R]: ...

else:
    join = either_t.join(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    def bound[L, From, To](either_value: Either[L, From], function: Callable[[From], Either[L, To]]) -> Either[L, To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[L, From, Intermediate, To](
        value: From,
        first: Callable[[From], Either[L, Intermediate]],
        second: Callable[[Intermediate], Either[L, To]],
    ) -> Either[L, To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    @curry
    def default[L, R](either_value: Either[L, R], default_value: R) -> R: ...

else:
    default = either_t.default(felis.identity.identity)(felis.identity.bind)


@curry
@curry
def catch[E: BaseException, From, To](value: From, function: Callable[[From], To], exception_type: type[E]) -> Either[E, To]:
    try:
        return Right(function(value))
    except exception_type as exception:
        return Left(exception)



from felis.predicate import comparison, interval

__all__ = ["negative", "non_negative", "non_positive", "non_zero", "portion", "positive"]


negative = comparison.less_than(0.0)


non_positive = comparison.less_than_or_same_as(0.0)


non_zero = comparison.different(0.0)


non_negative = comparison.greater_than_or_same_as(0.0)


positive = comparison.greater_than(0.0)


portion = interval.inclusive(0.0)(1.0)



from collections.abc import Callable

from felis.currying import curry

__all__ = ["flip"]


@curry
@curry
def flip[First, Second, Result](first: First, second: Second, function: Callable[[First], Callable[[Second], Result]]) -> Result:
    return function(first)(second)



from felis.bool import both


def test_both_returns_false_when_first_is_false():
    assert both(True)(False) is False


def test_both_returns_false_when_second_is_false():
    assert both(False)(True) is False


def test_both_returns_true_when_both_are_true():
    assert both(True)(True) is True



import math
from typing import NewType

import felis.predicate.float
from felis.constrained import smart_constructor

__all__ = [
    "Finite",
    "Negative",
    "NonNegative",
    "NonPositive",
    "NonZero",
    "Portion",
    "Positive",
    "finite",
    "negative",
    "non_negative",
    "non_positive",
    "non_zero",
    "portion",
    "positive",
]


Negative = NewType("Negative", float)

negative = smart_constructor(Negative)(felis.predicate.float.negative)


NonPositive = NewType("NonPositive", float)

non_positive = smart_constructor(NonPositive)(felis.predicate.float.non_positive)


NonZero = NewType("NonZero", float)

non_zero = smart_constructor(NonZero)(felis.predicate.float.non_zero)


NonNegative = NewType("NonNegative", float)

non_negative = smart_constructor(NonNegative)(felis.predicate.float.non_negative)


Positive = NewType("Positive", float)

positive = smart_constructor(Positive)(felis.predicate.float.positive)


Portion = NewType("Portion", float)

portion = smart_constructor(Portion)(felis.predicate.float.portion)


Finite = NewType("Finite", float)

finite = smart_constructor(Finite)(math.isfinite)



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, function, monad, state_t
from felis.currying import curry
from felis.lazy import Lazy

__all__ = [
    "ReversedState",
    "State",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "reversed_apply",
    "reversed_bind",
    "reversed_bound",
    "reversed_compose",
    "reversed_discard_after",
    "reversed_discard_before",
    "reversed_join",
    "reversed_lift2",
    "reversed_take_after",
    "reversed_take_before",
    "run",
    "take_after",
    "take_before",
    "when",
]


type State[S, T] = Callable[[S], tuple[T, S]]


type ReversedState[S, T] = State[Lazy[S], T]


if TYPE_CHECKING:

    @curry
    def run[S, T](state: S, state_value: State[S, T]) -> T: ...

else:
    run = state_t.run(felis.identity.map)


if TYPE_CHECKING:

    @curry
    @curry
    def map[S, From, To](state: S, state_value: State[S, From], function: Callable[[From], To]) -> tuple[To, S]: ...

else:
    map = state_t.map(felis.identity.map)


if TYPE_CHECKING:

    @curry
    def identity[S, T](state: S, value: T) -> tuple[T, S]: ...

else:
    identity = state_t.identity(felis.identity.identity)


if TYPE_CHECKING:

    @curry
    @curry
    def apply[S, From, To](state: S, state_value: State[S, From], state_function: State[S, Callable[[From], To]]) -> tuple[To, S]: ...

else:
    apply = state_t.apply(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[S, First, Second, Result](
        second: State[S, Second],
        first: State[S, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> State[S, Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when[S](bool: bool, state_none: State[S, None]) -> State[S, None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    @curry
    def join[S, T](state: S, state_state_value: State[S, State[S, T]]) -> tuple[T, S]: ...

else:
    join = state_t.join(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    def bound[S, From, To](state_value: State[S, From], function: Callable[[From], State[S, To]]) -> State[S, To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[S, From, Intermediate, To](
        value: From,
        first: Callable[[From], State[S, Intermediate]],
        second: Callable[[Intermediate], State[S, To]],
    ) -> State[S, To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    @curry
    @curry
    def reversed_apply[S, From, To](
        state: Lazy[S],
        state_value: ReversedState[S, From],
        state_function: ReversedState[S, Callable[[From], To]],
    ) -> tuple[To, Lazy[S]]: ...

else:
    reversed_apply = state_t.reversed_apply(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def reversed_lift2[S, First, Second, Result](
        second: ReversedState[S, Second],
        first: ReversedState[S, First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> ReversedState[S, Result]: ...

else:
    reversed_lift2 = applicative.lift2(map)(reversed_apply)


reversed_take_after = reversed_lift2(function.flip(function.identity))


reversed_discard_after = reversed_lift2(function.identity)


reversed_take_before = function.flip(reversed_discard_after)


reversed_discard_before = function.flip(reversed_take_after)


if TYPE_CHECKING:

    @curry
    def reversed_join[S, T](state: Lazy[S], state_state_value: ReversedState[S, ReversedState[S, T]]) -> tuple[T, Lazy[S]]: ...

else:
    reversed_join = state_t.reversed_join(felis.identity.identity)(felis.identity.bind)


if TYPE_CHECKING:

    @curry
    def reversed_bound[S, From, To](state_value: ReversedState[S, From], function: Callable[[From], ReversedState[S, To]]) -> ReversedState[S, To]: ...

else:
    reversed_bound = monad.bound(map)(reversed_join)


reversed_bind = function.flip(reversed_bound)


if TYPE_CHECKING:

    @curry
    @curry
    def reversed_compose[S, From, Intermediate, To](
        value: From,
        first: Callable[[From], ReversedState[S, Intermediate]],
        second: Callable[[Intermediate], ReversedState[S, To]],
    ) -> ReversedState[S, To]: ...

else:
    reversed_compose = monad.compose(reversed_bind)



from typing import NewType

import felis.predicate.float
from felis.constrained import smart_constructor

__all__ = ["Negative", "NonNegative", "NonPositive", "NonZero", "Positive", "negative", "non_negative", "non_positive", "non_zero", "positive"]


Negative = NewType("Negative", int)

negative = smart_constructor(Negative)(felis.predicate.float.negative)


NonPositive = NewType("NonPositive", int)

non_positive = smart_constructor(NonPositive)(felis.predicate.float.non_positive)


NonZero = NewType("NonZero", int)

non_zero = smart_constructor(NonZero)(felis.predicate.float.non_zero)


NonNegative = NewType("NonNegative", int)

non_negative = smart_constructor(NonNegative)(felis.predicate.float.non_negative)


Positive = NewType("Positive", int)

positive = smart_constructor(Positive)(felis.predicate.float.positive)



from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import felis.list
import felis.option
from felis import applicative, function, monad, option_t, state_t
from felis.currying import curry
from felis.option import Option
from felis.predicate import Predicate

__all__ = [
    "Parser",
    "add",
    "alnum",
    "alpha",
    "any",
    "apply",
    "bind",
    "bound",
    "bracket",
    "chain_left",
    "chain_left_1",
    "chain_right",
    "chain_right_1",
    "character",
    "compose",
    "digit",
    "discard_after",
    "discard_before",
    "end",
    "guard",
    "identity",
    "join",
    "lift2",
    "many",
    "map",
    "neutral",
    "option",
    "run",
    "satisfy",
    "separated",
    "some",
    "take_after",
    "take_before",
    "text",
    "when",
]


type Parser[T] = Callable[[str], Option[tuple[T, str]]]


if TYPE_CHECKING:

    @curry
    def run[T](string: str, parser_value: Parser[T]) -> Option[T]: ...

else:
    run = state_t.run(felis.option.map)


if TYPE_CHECKING:
    # [T : *] -> Parser T
    def neutral(string: str, /) -> Option[tuple[Any, str]]: ...

else:
    neutral = option_t.neutral(function.identity)


if TYPE_CHECKING:

    @curry
    def add[T](first: Parser[T], second: Parser[T]) -> Parser[T]: ...

else:
    add = option_t.add(function.bind)


if TYPE_CHECKING:

    @curry
    def map[From, To](parser_value: Parser[From], function: Callable[[From], To]) -> Parser[To]: ...

else:
    map = state_t.map(felis.option.map)


if TYPE_CHECKING:

    def identity[T](value: T, /) -> Parser[T]: ...

else:
    identity = state_t.identity(felis.option.identity)


if TYPE_CHECKING:

    @curry
    def apply[From, To](parser_value: Parser[From], parser_function: Parser[Callable[[From], To]]) -> Parser[To]: ...

else:
    apply = state_t.apply(felis.option.identity)(felis.option.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: Parser[Second],
        first: Parser[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Parser[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, parser_none: Parser[None]) -> Parser[None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T](parser_value: Parser[Parser[T]], /) -> Parser[T]: ...

else:
    join = state_t.join(felis.option.identity)(felis.option.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](parser_value: Parser[From], function: Callable[[From], Parser[To]]) -> Parser[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], Parser[Intermediate]],
        second: Callable[[Intermediate], Parser[To]],
    ) -> Parser[To]: ...

else:
    compose = monad.compose(bind)


if TYPE_CHECKING:

    def guard(bool: bool) -> Parser[None]: ...

else:
    guard = monad.guard(neutral)(identity)


def end(string: str) -> Option[tuple[None, str]]:
    return None if string else felis.option.Some((None, ""))


def any(string: str) -> Option[tuple[str, str]]:
    return felis.option.Some((string[0], string[1:])) if string else None


def satisfy(predicate: Predicate[str]) -> Parser[str]:
    return bind(any)(lambda character: identity(character) if predicate(character) else neutral)


def character(character: str) -> Parser[str]:
    return satisfy(lambda current: current == character)


def text(string: str) -> Parser[str]:
    return take_after(character(string[0]))(take_after(text(string[1:]))(identity(string))) if string else identity("")


def many[T](parser: Parser[T]) -> Parser[list[T]]:
    return add(identity(felis.list.neutral))(bind(parser)(lambda first: bind(many(parser))(lambda rest: identity([first, *rest]))))


def some[T](parser: Parser[T]) -> Parser[list[T]]:
    return bind(parser)(lambda first: bind(many(parser))(lambda rest: identity([first, *rest])))


def option[T](parser: Parser[T]) -> Parser[Option[T]]:
    return add(identity(felis.option.neutral))(map(felis.option.identity)(parser))


# [S : *] -> Parser S -> [T : *] -> Parser T -> Parser (list T)
@curry
def separated[T](parser: Parser[T], separator: Parser[Any]) -> Parser[list[T]]:
    return add(identity(felis.list.neutral))(bind(parser)(lambda first: bind(many(take_after(separator)(parser)))(lambda rest: identity([first, *rest]))))


# [L : *] -> Parser L -> [R : *] -> Parser R -> [T : *] -> Parser T -> Parser T
@curry
@curry
def bracket[T](parser: Parser[T], right: Parser[Any], left: Parser[Any]) -> Parser[T]:
    return take_after(left)(take_before(right)(parser))


@curry
@curry
def chain_right[R, T](parser_value: Parser[T], parser_function: Parser[Callable[[T], Callable[[R], R]]], accumulator: R) -> Parser[R]:
    return add(identity(accumulator))(
        bind(parser_function)(
            lambda function: bind(parser_value)(
                lambda value: bind(chain_right(accumulator)(parser_function)(parser_value))(lambda accumulator: identity(function(value)(accumulator))),
            ),
        ),
    )


@curry
def chain_right_1[T](parser_value: Parser[T], parser_function: Parser[Callable[[T], Callable[[T], T]]]) -> Parser[T]:
    def rest(accumulator: T) -> Parser[T]:
        return add(identity(accumulator))(
            bind(parser_function)(lambda function: bind(parser_value)(lambda value: bind(rest(value))(lambda value: identity(function(accumulator)(value))))),
        )

    return bind(parser_value)(rest)


@curry
@curry
def chain_left[R, T](parser_value: Parser[T], parser_function: Parser[Callable[[R], Callable[[T], R]]], accumulator: R) -> Parser[R]:
    return add(identity(accumulator))(
        bind(parser_function)(lambda function: bind(parser_value)(lambda value: chain_left(function(accumulator)(value))(parser_function)(parser_value))),
    )


@curry
def chain_left_1[T](parser: Parser[T], function: Parser[Callable[[T], Callable[[T], T]]]) -> Parser[T]:
    return bind(parser)(lambda first: chain_left(first)(function)(parser))


digit = satisfy(str.isdigit)


alpha = satisfy(str.isalpha)


alnum = satisfy(str.isalnum)



from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from felis.currying import curry

__all__ = ["Either", "Left", "Right", "add", "apply", "default", "join"]


type Either[L, R] = Left[L] | Right[R]


@dataclass(frozen=True)
class Left[T]:
    value: T


@dataclass(frozen=True)
class Right[T]:
    value: T


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [L : *] -> [R : *] -> M (Either L R) -> M (Either L R) -> M (Either L R)
@curry
@curry
@curry
def add(m_augend: Any, m_addend: Any, m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]], m_identity: Callable[[Any], Any]) -> Any:
    def augend_binder(augend: Either[Any, Any]) -> Any:
        match augend:
            case Left(_):
                return m_addend
            case Right(right):
                return m_identity(Right(right))

    return m_bind(m_augend)(augend_binder)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [L : *] -> [From : *] -> [To : *] -> M (Either L (From -> To)) -> M (Either L From) -> M (Either L To)
@curry
@curry
@curry
def apply(
    m_either_value: Any,
    m_either_function: Any,
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def either_function_binder(either_function: Either[Any, Any]) -> Any:
        match either_function:
            case Left(left):
                return m_identity(Left(left))
            case Right(function):

                def either_value_binder(either_value: Either[Any, Any]) -> Any:
                    match either_value:
                        case Left(left):
                            return m_identity(Left(left))
                        case Right(value):
                            return m_identity(Right(function(value)))

                return m_bind(m_either_value)(either_value_binder)

    return m_bind(m_either_function)(either_function_binder)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [L : *] -> [R : *] -> M (Either L (M (Either L R))) -> M (Either L R)
@curry
@curry
def join(m_either_m_either_value: Any, m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]], m_identity: Callable[[Any], Any]) -> Any:
    def either_m_either_binder(either_m_either_value: Either[Any, Any]) -> Any:
        match either_m_either_value:
            case Left(left):
                return m_identity(Left(left))
            case Right(m_either_value):
                return m_either_value

    return m_bind(m_either_m_either_value)(either_m_either_binder)


# [M : * -> *] ->
# ([T : *] -> T -> M T) ->
# ([From : *] -> [To : *] -> M From -> (From -> M To) -> M To) ->
# [L : *] -> [R : *] -> M R -> M (Either L R) -> M R
@curry
@curry
@curry
def default(
    m_either_value: Any,
    default_value: Any,
    m_bind: Callable[[Any], Callable[[Callable[[Any], Any]], Any]],
    m_identity: Callable[[Any], Any],
) -> Any:
    def either_binder(either_value: Either[Any, Any]) -> Any:
        match either_value:
            case Left(_):
                return default_value
            case Right(value):
                return m_identity(value)

    return m_bind(m_either_value)(either_binder)



from collections.abc import Callable
from typing import TYPE_CHECKING

import felis.identity
from felis import applicative, coroutine, function, lazy, lazy_t, monad
from felis.coroutine import Coroutine
from felis.currying import curry
from felis.lazy import Lazy

__all__ = [
    "LazyCoroutine",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


type LazyCoroutine[T] = Lazy[Coroutine[T]]


if TYPE_CHECKING:

    @curry
    def map[From, To](lazy_coroutine_value: LazyCoroutine[From], function: Callable[[From], To]) -> LazyCoroutine[To]: ...

else:
    map = felis.identity.compose(lazy.map)(coroutine.map)


identity = felis.identity.compose(lazy.identity)(coroutine.identity)


lift = lazy.map(coroutine.identity)


if TYPE_CHECKING:

    @curry
    @curry
    def apply[From, To](lazy_coroutine_value: LazyCoroutine[From], lazy_coroutine_function: LazyCoroutine[Callable[[From], To]]) -> LazyCoroutine[To]: ...

else:
    apply = lazy_t.apply(coroutine.identity)(coroutine.bind)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: LazyCoroutine[Second],
        first: LazyCoroutine[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> LazyCoroutine[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, lazy_coroutine_none: LazyCoroutine[None]) -> LazyCoroutine[None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T](lazy_coroutine_lazy_coroutine_value: LazyCoroutine[LazyCoroutine[T]]) -> LazyCoroutine[T]: ...

else:
    join = lazy_t.join(coroutine.identity)(coroutine.bind)


if TYPE_CHECKING:

    @curry
    def bound[From, To](lazy_coroutine_value: LazyCoroutine[From], function: Callable[[From], LazyCoroutine[To]]) -> LazyCoroutine[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], LazyCoroutine[Intermediate]],
        second: Callable[[Intermediate], LazyCoroutine[To]],
    ) -> LazyCoroutine[To]: ...

else:
    compose = monad.compose(bind)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import felis.identity
from felis import applicative, free_t, function, monad
from felis.currying import curry
from felis.free_t import Pure

__all__ = [
    "Bind",
    "Free",
    "Pure",
    "apply",
    "bind",
    "bound",
    "compose",
    "discard_after",
    "discard_before",
    "identity",
    "join",
    "lift2",
    "map",
    "take_after",
    "take_before",
    "when",
]


if TYPE_CHECKING:

    @dataclass(frozen=True)
    class Bind[T]:
        f_free_value: felis.identity.Identity[Free[T]]

else:
    from felis.free_t import Bind


if TYPE_CHECKING:
    type Free[T] = Pure[T] | Bind[T]
else:
    from felis.free_t import Free


if TYPE_CHECKING:

    @curry
    def map[From, To](free_value: Free[From], function: Callable[[From], To]) -> Free[To]: ...

else:
    map = free_t.map(felis.identity.map)


if TYPE_CHECKING:
    # [T : *] -> T -> Free T
    identity: Free[Any]
else:
    from felis.free_t import identity


if TYPE_CHECKING:

    @curry
    def apply[From, To](free_value: Free[From], free_function: Free[Callable[[From], To]]) -> Free[To]: ...

else:
    apply = free_t.apply(felis.identity.map)


if TYPE_CHECKING:

    @curry
    @curry
    def lift2[First, Second, Result](
        second: Free[Second],
        first: Free[First],
        function: Callable[[First], Callable[[Second], Result]],
    ) -> Free[Result]: ...

else:
    lift2 = applicative.lift2(map)(apply)


take_after = lift2(function.flip(function.identity))


discard_after = lift2(function.identity)


take_before = function.flip(discard_after)


discard_before = function.flip(take_after)


if TYPE_CHECKING:

    @curry
    def when(bool: bool, free_none: Free[None]) -> Free[None]: ...

else:
    when = applicative.when(identity)


if TYPE_CHECKING:

    def join[T](free_free_value: Free[Free[T]], /) -> Free[T]: ...

else:
    join = free_t.join(felis.identity.map)


if TYPE_CHECKING:

    @curry
    def bound[From, To](free_value: Free[From], function: Callable[[From], Free[To]]) -> Free[To]: ...

else:
    bound = monad.bound(map)(join)


bind = function.flip(bound)


if TYPE_CHECKING:

    @curry
    @curry
    def compose[From, Intermediate, To](
        value: From,
        first: Callable[[From], Free[Intermediate]],
        second: Callable[[Intermediate], Free[To]],
    ) -> Free[To]: ...

else:
    compose = monad.compose(bind)



from felis import function
from felis.currying import curry

__all__ = ["Bool", "both", "different", "either", "negate", "same"]


Bool = bool


@curry
def both(first: bool, second: bool) -> bool:
    return first and second


@curry
def either(first: bool, second: bool) -> bool:
    return first or second


def negate(value: bool) -> bool:
    return not value


@curry
def same(first: bool, second: bool) -> bool:
    return first is second


different = function.map2(negate)(same)



from felis.bool import Bool
from felis.coroutine import Coroutine
from felis.coroutine_option import CoroutineOption
from felis.dict import Dict
from felis.either import Either
from felis.free import Free
from felis.free_dict import FreeDict
from felis.function import Function
from felis.function_option import FunctionOption
from felis.identity import Identity
from felis.iterable import Iterable
from felis.lazy import Lazy
from felis.lazy_coroutine import LazyCoroutine
from felis.lazy_coroutine_either import LazyCoroutineEither
from felis.lazy_coroutine_either_list import LazyCoroutineEitherList
from felis.lazy_coroutine_option import LazyCoroutineOption
from felis.list import List
from felis.option import Option
from felis.order import Order
from felis.ordering import Ordering
from felis.pair import Pair
from felis.parser import Parser
from felis.predicate import Predicate
from felis.state import ReversedState, State

__all__ = [
    "Bool",
    "Coroutine",
    "CoroutineOption",
    "Dict",
    "Either",
    "Free",
    "FreeDict",
    "Function",
    "FunctionOption",
    "Identity",
    "Iterable",
    "Lazy",
    "LazyCoroutine",
    "LazyCoroutineEither",
    "LazyCoroutineEitherList",
    "LazyCoroutineOption",
    "List",
    "Option",
    "Order",
    "Ordering",
    "Pair",
    "Parser",
    "Predicate",
    "ReversedState",
    "State",
]
